using namespace std;
/*************************************************************** 
* 
* HiLowArrayData.cc             & XMM Science Analysis System (SAS) 
* 
* Class representing array data object.
* 
* (c) 1997 ESA - Design by Marco Beijersbergen
*                Written by Mark Thomas, CARA
* 
* $Id: HiLowArrayData.cc,v 1.2 2004/03/05 14:43:02 eojero Exp $
* 
*****************************************************************/ 

#include <string>
#include "errstr.h"
#include "Tracer.h"
#include "HiLowArrayData.h"
#include "Storage.h"

#include <cassert>

static unsigned long
numberOfElements( const longVec & vec )
{
	PROC

	unsigned long	elements = 1;

	for( unsigned int i = 0; i < vec.size(); ++i ) elements *= vec[ i ];

	RETURN elements;
}


HiLowArrayData::HiLowArrayData(
			HiLowDataStorage *		data,
			const DalArray *		parent,
			const longVec&			from,
			const longVec&			count,
			AccessMode			accessMode )
		:	DalArrayData( parent, from, count ),
			HiLowData<DalArray>( data, 0, 0, parent, accessMode )
{
	PROC

	seek( from, count );

	ENDPROC
}


HiLowArrayData::~HiLowArrayData()
{
	PROC

	((Storage<HiLowArrayData>&)*dataStorage()).parent( this );

	unsigned long c = 1;
	for( unsigned int i = 0; i < DalSeekable<longVec>::count().size(); ++i )
	{
		c *= DalSeekable<longVec>::count()[ i ];
	} // for
	dataStorage() -> release( position( from() ), c );

	DalArray* array = this -> DalChild<DalArray>::parent();
	array -> removeDataObject( this );

	ENDPROC
}


unsigned long
HiLowArrayData::elements() const
{
	PROC
	RETURN this -> DalChild<Array>::parent() -> elements();
}


const longVec &
HiLowArrayData::size() const
{
	PROC
	RETURN this -> DalChild<Array>::parent() -> size();
}


void
HiLowArrayData::seek(
	const longVec & from,
	const longVec & count )
{
	PROC

	//dataStorage() -> release( seekable().from(), seekable().count() );

	const Array * array = this -> DalChild<Array>::parent();

	const longVec& size = array->size();
	longVec _from;
	longVec _count;

	if( from.size() == 0 )
	{
		_from = size;
		for (unsigned i = 0; i < _from.size(); ++i)
			_from[i] = 0;
	}
	else if( from.size() == size.size() )
	{
		_from = from;
	}
	else
	{
		errstr << "could not access the data in array " << array->parent()->name() << "(" << array->name() << ") since the dimension " << from.size() << " of the starting point does not match the dimension " << array->dimensions() << " of the column" << Sas::error("invalidFromDimensions");

	} // if

	if ( count.size() == 0 )
	{
		_count = size;
	}
	else if ( count.size() == size.size() )
	{
		_count = count;
	}
	else
	{
		errstr << "could not access the data in array " << array->parent()->name() << "(" << array->name() << ") since the dimension " << count.size() << " of the range of interest does not match the dimension " << array->dimensions() << " of the column" << Sas::error("invalidCountDimensions");
	} // if

	for( unsigned i = 0; i < _from.size(); ++i )
	{
		if (_from[i] + _count[i] > size[i])
		{
			errstr << "could not access the data in array " << array->parent()->name() << "(" << array->name() << ") since the range of interest " << _from[i] << " .. " << _from[i] + _count[i] << " goes outside its size " << size[i] << Sas::error("invalidArrayDataRange");
		} // if
	} // for

	DalSeekable< longVec >::seek( _from, _count );

	assert( DalSeekable< longVec >::from() == _from );
	assert(	DalSeekable< longVec >::count() == _count );

	DalSeekable<unsigned long> oldSeek = seekable();

	seekable().seek(
			position( _from ),
			::numberOfElements( _count ));

	dataStorage() -> initialise( oldSeek, seekable() );

	ENDPROC
}


template<class T>
static void dumpData(
	const HiLowArrayData *	arrayData,
	T			values,
	unsigned long 		from,
	unsigned long 		count )
{
	PROC

	const DalArray * array = arrayData -> DalChild<DalArray>::parent();

	if( arrayData -> mode() == Dal::Read )
	{
		// ***************
		// General Message
		// ***************
		if( errHandler.verbosity(Msg::App) >= Msg::Verbose )
		{
			errstr << "Data for array with qualified name " << array -> qualifiedName() << " from element " << from << " to element " << from + count - 1 << ", is read-only, and will not be written back to disk." << Sas::message( Msg::Core, Msg::Verbose );
		} // if

		ALTRETURN;
	} // if

	const DalDataSet * dataSet = ((DalChild<DalDataSet>&)*array).parent();
        dataSet -> dataSetReaderWriter() -> dumpData( *array, values, from, count );
	ENDPROC
}


template<class T>
static void loadData(
	const HiLowArrayData *	arrayData,
	T			values,
	unsigned long 		from,
	unsigned long 		count )
{
	PROC
	const DalArray * array = arrayData -> DalChild<DalArray>::parent();
	const DalDataSet * dataSet = ((DalChild<DalDataSet>&)*array).parent();
        dataSet -> dataSetReaderWriter() -> loadData( *array, values, from, count );
	ENDPROC
}


void
HiLowArrayData::dumpData( char ** values, unsigned long from, unsigned long count ) const
{
	PROC
	::dumpData( this, values, from, count );
	ENDPROC
}


void
HiLowArrayData::dumpData( bool8 * values, unsigned long from, unsigned long count ) const
{
	PROC
	::dumpData( this, values, from, count );
	ENDPROC
}


void
HiLowArrayData::dumpData( int8 * values, unsigned long from, unsigned long count ) const
{
	PROC
	::dumpData( this, values, from, count );
	ENDPROC
}


void
HiLowArrayData::dumpData( uint16 * values, unsigned long from, unsigned long count ) const
{
	PROC
	::dumpData( this, values, from, count );
	ENDPROC
}


void
HiLowArrayData::dumpData( int16 * values, unsigned long from, unsigned long count ) const
{
	PROC
	::dumpData( this, values, from, count );
	ENDPROC
}


void
HiLowArrayData::dumpData( uint32 * values, unsigned long from, unsigned long count ) const
{
	PROC
	::dumpData( this, values, from, count );
	ENDPROC
}


void
HiLowArrayData::dumpData( int32 * values, unsigned long from, unsigned long count ) const
{
	PROC
	::dumpData( this, values, from, count );
	ENDPROC
}


void
HiLowArrayData::dumpData( real32 * values, unsigned long from, unsigned long count ) const
{
	PROC
	::dumpData( this, values, from, count );
	ENDPROC
}


void
HiLowArrayData::dumpData( real64 * values, unsigned long from, unsigned long count ) const
{
	PROC
	::dumpData( this, values, from, count );
	ENDPROC
}


void
HiLowArrayData::loadData( char ** values, unsigned long from, unsigned long count ) const
{
	PROC
	::loadData( this, values, from, count );
	ENDPROC
}


void
HiLowArrayData::loadData( bool8 * values, unsigned long from, unsigned long count ) const
{
	PROC
	::loadData( this, values, from, count );
	ENDPROC
}


void
HiLowArrayData::loadData( int8 * values, unsigned long from, unsigned long count ) const
{
	PROC
	::loadData( this, values, from, count );
	ENDPROC
}


void
HiLowArrayData::loadData( uint16 * values, unsigned long from, unsigned long count ) const
{
	PROC
	::loadData( this, values, from, count );
	ENDPROC
}


void
HiLowArrayData::loadData( int16 * values, unsigned long from, unsigned long count ) const
{
	PROC
	::loadData( this, values, from, count );
	ENDPROC
}


void
HiLowArrayData::loadData( uint32 * values, unsigned long from, unsigned long count ) const
{
	PROC
	::loadData( this, values, from, count );
	ENDPROC
}


void
HiLowArrayData::loadData( int32 * values, unsigned long from, unsigned long count ) const
{
	PROC
	::loadData( this, values, from, count );
	ENDPROC
}


void
HiLowArrayData::loadData( real32 * values, unsigned long from, unsigned long count ) const
{
	PROC
	::loadData( this, values, from, count );
	ENDPROC
}


void
HiLowArrayData::loadData( real64 * values, unsigned long from, unsigned long count ) const
{
	PROC
	::loadData( this, values, from, count );
	ENDPROC
}


void
HiLowArrayData::copyData( unsigned long , unsigned long , unsigned long ) const
//  HiLowArrayData::copyData( unsigned long from, unsigned long to, unsigned long count ) const
{
	PROC
	errstr << "HiLowArrayData::copyData" << Sas::error("notImplemented" );
	ENDPROC
}


int8 *
HiLowArrayData::int8Data() const
{
	PROC
	((Storage<HiLowArrayData>&)*dataStorage()).parent( this );
	RETURN dataStorage() -> int8Data( seekable().from(), seekable().count() );
}


uint16 *
HiLowArrayData::uint16Data() const
{
	PROC
	((Storage<HiLowArrayData>&)*dataStorage()).parent( this );
	RETURN dataStorage() -> uint16Data( seekable().from(), seekable().count() );
}


int16 *
HiLowArrayData::int16Data() const
{
	PROC
	((Storage<HiLowArrayData>&)*dataStorage()).parent( this );
	RETURN dataStorage() -> int16Data( seekable().from(), seekable().count() );
}


uint32 *
HiLowArrayData::uint32Data() const
{
	PROC
	((Storage<HiLowArrayData>&)*dataStorage()).parent( this );
	RETURN dataStorage() -> uint32Data( seekable().from(), seekable().count() );
}


int32 *
HiLowArrayData::int32Data() const
{
	PROC
	((Storage<HiLowArrayData>&)*dataStorage()).parent( this );
	RETURN dataStorage() -> int32Data( seekable().from(), seekable().count() );
}


real32 *
HiLowArrayData::real32Data() const
{
	PROC
	((Storage<HiLowArrayData>&)*dataStorage()).parent( this );
	RETURN dataStorage() -> real32Data( seekable().from(), seekable().count() );
}


real64 *
HiLowArrayData::real64Data() const
{
	PROC
	((Storage<HiLowArrayData>&)*dataStorage()).parent( this );
	RETURN dataStorage() -> real64Data( seekable().from(), seekable().count() );
}


bool8 *
HiLowArrayData::boolData() const
{
	PROC
	((Storage<HiLowArrayData>&)*dataStorage()).parent( this );
	RETURN dataStorage() -> boolData( seekable().from(), seekable().count() );
}


char *
HiLowArrayData::stringData() const
{
	PROC
	((Storage<HiLowArrayData>&)*dataStorage()).parent( this );
	RETURN dataStorage() -> stringData( seekable().from(), seekable().count() );
}

