using namespace std;
/*************************************************************** 
 * 
 * FitsReaderWriter.cc     & XMM Science Analysis System (SAS) 
 * 
 * Class representing FITS file reader/writer.
 * 
 * (c) 1997 ESA - Written by Mark Thomas, CARA
 * 
 * $Id: FitsReaderWriter.cc,v 1.4 2012/05/23 16:19:04 aibarra Exp $ 
 * 
 *****************************************************************/ 

#include <fitsio.h>
#include <fstream>
#include <sstream>

#include <STime.h>
#include <stringutil.h>
#include <Tracer.h>

#include "FitsReaderWriter.h"
#include "dal_utilities.h"

#include<ExceptionErrorHandler.h>


// ********************
// Global defined here.
// ********************
const string FitsReaderWriter::attributeSeparator = "XMMSEP";
const string FitsReaderWriter::attributeSeparatorComment = " primary image keywords follow";
const string FitsReaderWriter::defaultColumnComment = "The name of this column";
const string FitsReaderWriter::arrayUnitsComment = "Units of the image";
const string FitsReaderWriter::columnNullValueComment = "Null value of the column";
const string FitsReaderWriter::arrayNullValueComment = "Null value of the array";

// ===================================================================
// Some version of cfitsio prepend COMMENT and HISTORY records with
// two spaces. If such a version is being used, the following constant
// should be set to 2, otherwise, it should be set to zero. This hard
// coding could be replaced with mechanism to read this value
// from a configuration file or some other mechanism.
// ===================================================================
static const unsigned numberOfFITSSpaces = 0;


// *********************************************************************
// The appendThisString argument will be appended to the attribute name.
// *********************************************************************
int
FitsReaderWriter::writeAttributes( 
    fitsfile *		fptr,
    const Attributable &	attributeSet,
    const string &		appendThisString,
    bool (FitsReaderWriter::*attributeOK)(const string &) const,
    int &			status ) const
{

    PROC

	if( status != 0 )
	{
	    ALTRETURNV( 1 );
	} // if


    // ***************
    // write keywords.
    // ***************
    for(
	unsigned int i = 0;
	i < attributeSet.attributes();
	++i )
    {
	const Attribute & attribute = *attributeSet.attribute( i );

	// ==============================================
	// Construct the attribute, which will be written
	// to the Fits file.
	// ==============================================
	const string attributeName =
	    attribute.name()
	    + appendThisString;


	// ====================================================
	// Any Attribute names which are not Fits compatible
	// will be ignored (a warning message will be displayed
	// to indicate this).
	// The following will be ignored:
	// - FITS reserved keywords
	// - Names with more than 8 characters (this is the
	//   safest course of action to take as this
	//   will prevent a backdoor way of setting
	//   reserved keywords.
	// ====================================================
	if( isFITSReservedKeyword( attributeName ))
	{
	    errstr << "The attribute with name '" << attributeName << "' is a FITS reserved keyword and is not being written to the Fits file." << Sas::warning( "ReservedKeyword" );

	    continue;
	} // if

	if( attributeName.length() > 8 )
	{
	    errstr << "The attribute with name '" << attributeName << "' is too long and is not being written to the Fits file." << Sas::warning( "KeywordTooLong" );
	    continue;
	} // if


	// ==============================================
	// Do some additional checks, which are dependant
	// on the type of object the attribute belongs
	// to.
	// ==============================================
	//if( attributeOK && !attributeOK( attribute.name() ))
	if( attributeOK && !(this ->* attributeOK)( attribute.name() ))
	{
	    continue;
	} // if


	// ==========================================
	// Everything is OK so write the attribute to
	// the Fits file.
	// ==========================================
	switch ( attribute.dataType() )
	{
	    case Attribute::Bool:
		fits_update_key_log(
		    fptr,
		    (char*)attributeName.c_str(),
		    attribute.asBool(),
		    (char*)attribute.label().c_str(),
		    &status);

		break;

	    case Attribute::Int:
		fits_update_key_lng(
		    fptr,
		    (char*)attributeName.c_str(),
		    attribute.asInt(),
		    (char*)attribute.label().c_str(),
		    &status);

		break;

	    case Attribute::Real:
		fits_update_key_dbl(
		    fptr,
		    (char*)attributeName.c_str(),
		    attribute.asReal(),
		    14,
		    (char*)attribute.label().c_str(),
		    &status);

		break;

	    case Attribute::DString:

		// ********************************************
		// To prevent long strings from being
		// duplicated it is necessary to
		// delete the current keyword (if it
		// exists manually, since fits does not provide
		// the function fits_update_longstr().
		// ********************************************
		//fits_delete_key(
		//	fptr,
		//	(char*)attributeName.c_str(),
		//	&status );

		//if( status == KEY_NO_EXIST ) status = 0;
		//fits_check( fptr );

		fits_write_key_longstr(
		    fptr,
		    (char*)attributeName.c_str(),
		    (char*)attribute.asString().c_str(),
		    (char*)attribute.label().c_str(),
		    &status);

		break;

	    default:
		errstr << "could not write attribute '" << attribute.name() << "' of '" << attribute.parent()->name() << "' to file '" << fits_file_name( fptr ) << "' since its data type " << int( attribute.dataType() ) << " is not recognised" << Sas::error("invalidAttributeDataType");

		break;

	} // switch

	// =====================================
	// write units to keyword comment field.
	// =====================================
	if( !attribute.units().empty() )
	{
	    fits_write_key_unit(
		fptr,
		(char *)attributeName.c_str(),
		(char *)attribute.units().c_str(),
		&status );
	} // for
	fits_check(fptr);

    } // for

    // ******************************
    // Everything is Ok, so return 0.
    // ******************************

    RETURN 0;

}


int
FitsReaderWriter::writeComments( 
    fitsfile *		fptr,
    const Attributable &	attributeSet,
    int &			status ) const
{

    PROC

	if( status != 0 )
	{

	    ALTRETURNV( 1 );
	} // if

    for(
	unsigned int i = 0;
	i < attributeSet.comments();
	++i )
    {
	const string&	nextString = attributeSet.comment( i );

	if( isBlank( nextString ) )
	{
	    errstr << "A blank comment (for writing) has been detected in the attributable set with qualified name '" << attributeSet.qualifiedName() << "'. Since FITS does not handle these correctly, a string with exactly one space will be written." << Sas::message( Msg::Core, Msg::Verbose );

	    fits_write_comment(
		fptr,
		" ",	// At least one space is required,
		// otherwise cfitsio does not create the record
		&status );

	} 
	else
	{
	    fits_write_comment(
		fptr,
		nextString.c_str(),
		&status );

	} // if

	fits_check(fptr);
    } // for

    // ******************************
    // Everything is Ok, so return 0.
    // ******************************

    RETURN 0;
}


int
FitsReaderWriter::writeHistory( 
    fitsfile *		fptr,
    const Attributable &	attributeSet,
    int &			status ) const
{

    PROC

	if( status != 0 )
	{
	    ALTRETURNV( 1 );
	} // if

    for(
	unsigned int i = 0;
	i < attributeSet.historys();
	++i )
    {
	const string&	nextString = attributeSet.history( i );

	if( isBlank( nextString ) )
	{
	    errstr << "A blank history record (for writing) has been detected in the attributable set with qualified name '" << attributeSet.qualifiedName() << "'. Since FITS does not handle these correctly, a string with exactly one space will be written." << Sas::message( Msg::Core, Msg::Verbose );

	    fits_write_history(
		fptr,
		" ",	// At least one space is required,
		// otherwise cfitsio does not create the record
		&status );

	} 
	else
	{
	    fits_write_history(
		fptr,
		nextString.c_str(),
		&status );

	} // if

	fits_check(fptr);
    } // for

    // ******************************
    // Everything is Ok, so return 0.
    // ******************************

    RETURN 0;

}


int
FitsReaderWriter::writeHeader(
    fitsfile *		fptr,
    const Attributable &	attributeSet,
    const string &		appendThisString,
    bool (FitsReaderWriter::*attributeOK)(const string &) const,
    int &			status ) const
{

    PROC

	if( status != 0 )
	{
	    ALTRETURNV( 1 );
	} // if

    writeAttributes( fptr, attributeSet, appendThisString, attributeOK, status );

    writeComments( fptr, attributeSet, status );

    writeHistory( fptr, attributeSet, status );

    // ******************************
    // Everything is Ok, so return 0.
    // ******************************

    RETURN 0;
}


int
FitsReaderWriter::readCreatorAndDate(
    fitsfile*		fptr,
    const DataSet*		dataSet,
    int&			status ) const
{
    PROC

	if( status != 0 )
	{
	    ALTRETURNV( 1 );
	} // if

    char 	comment[ 1024 ];
    char*	value = 0;

    // CREATOR
    fits_read_key_longstr(
	fptr,
	"CREATOR",
	&value,
	comment,
	&status );

    // --------------------------
    // Check that CREATOR exists.
    // --------------------------
    if( !status )
    {
	dataSetServer -> creator( dataSet, value );
	free( value );
    }
    else if( status == KEY_NO_EXIST )
    {
	errstr << "The CREATOR keyword could not be found in dataset with name '" << dataSet -> name() << "'" << Sas::warning( "creatorKeywordNotFound" );

	dataSetServer -> creator( dataSet, "" );

	status = 0;

    } // if

    fits_check(fptr);

    // DATE
    fits_read_key_longstr(
	fptr,
	"DATE",
	&value,
	comment,
	&status );

    // -----------------------
    // Check that Date exists.
    // -----------------------
    if( !status )
    {
	dataSetServer -> date( dataSet, value );
	free( value );
    }
    else if( status == KEY_NO_EXIST )
    {
	errstr << "The DATE keyword could not be found in dataset with name '" << dataSet -> name() << "'" << Sas::warning( "dateKeywordNotFound" );

	dataSetServer -> date( dataSet, "" );

	status = 0;

    } // if

    fits_check(fptr);


    // ******************************
    // Everything is Ok, so return 0.
    // ******************************

    RETURN 0;
}


int
FitsReaderWriter::writeCreatorAndDate(
    fitsfile*		fptr,
    const DataSet*		dataSet,
    int&			status ) const
{
    PROC

	if( status != 0 )
	{
	    ALTRETURNV( 1 );
	} // if

    string creator;
    string date;

    if( dataSet -> mode() == Dal::Create )
    {
	creator = dataSetServer -> client();
	date = STime::currentTime().toString();
    }
    else
    {
	creator = dataSetServer -> creator( dataSet );
	date = dataSetServer -> date( dataSet );
    } // if

    fits_write_key_longstr(
	fptr,
	"CREATOR",
	(char*)( creator.c_str() ),
	"name of code that generated this file",
	&status );


    fits_write_key_longstr(
	fptr,
	"DATE",
	(char*)( date.c_str() ),
	"creation date",
	&status );

    fits_check(fptr);

    // ******************************
    // Everything is Ok, so return 0.
    // ******************************

    RETURN 0;
}


int
FitsReaderWriter::readProcessHistory(
    fitsfile*		fptr,
    const DataSet*		dataSet,
    int&			status ) const
{
    PROC

	if( status != 0 )
	{
	    ALTRETURNV( 1 );
	} // if

    int			processNumber = 0; // The first process has number 0.

    while( !status )
    {
	char 	comment[ 1024 ];
	char*	value;
	char	xproc[ 9 ]; // "XPROCnnn" + null

	sprintf( xproc, "XPROC%d", processNumber );

	fits_read_key_longstr(
	    fptr,
	    xproc,
	    &value,
	    comment,
	    &status );

	// ------------------------------------
	// Either, there are no more xproc
	// keywords
	// (KEY_NO_EXIST), in which case,
	// do not attempt to read any more
	// and exit from the loop, otherwise,
	// there was a real error.
	// ------------------------------------
	if( status == KEY_NO_EXIST )
	{
	    status = 0;

	    ALTRETURNV( 0 );

	} // if

	fits_check(fptr);

	dataSetServer -> process( dataSet, value );
	free( value );
	processNumber++;

    } // while

    // ******************************
    // If we get here an error occured.
    // ******************************

    RETURN 1;

}


int
FitsReaderWriter::writeProcessHistory(
    fitsfile*		fptr,
    const DataSet*		dataSet,
    int&			status ) const
{
    PROC

	if( status != 0 )
	{
	    ALTRETURNV( 1 );
	} // if

    char	xproc[ 9 ]; // "XPROCnnn" + null
    const unsigned int processes = dataSetServer -> processes( dataSet );

    for( unsigned int i = 0; i < processes; ++i )
    {
	sprintf( xproc, "XPROC%d", i );

	fits_write_key_longstr(
	    fptr,
	    xproc,
	    (char*)dataSetServer -> process( dataSet, i ).c_str(),
	    "",
	    //attributeName.c_str(),
	    //(char*)attribute.asString().c_str(),
	    //(char*)attribute.label().c_str(),
	    &status );
    } // for

    // Due to a bug in cfitsio earlier than 2.300 empty
    // stringattributes were not being written to disk. This bug
    // has now been fixed, so this test on the empty string here
    // is required in oder to avoid a number of spurious empty
    // XPROCs appearing in the output. This problem was caught by
    // the test harness process.cc.
    if(!dataSetServer->process().empty()){ 
	sprintf( xproc, "XPROC%d", processes );

	fits_write_key_longstr(
	    fptr,
	    xproc,
	    (char*)( dataSetServer -> process().c_str() ),
	    "",
	    &status );
    }
    // ******************************
    // Everything is Ok, so return 0.
    // ******************************

    RETURN 0;
}


int
FitsReaderWriter::readDebugHistory(
    fitsfile*		fptr,
    const DataSet*		dataSet,
    int&			status ) const
{
    PROC

	if( status != 0 )
	{
	    ALTRETURNV( 1 );
	} // if

    int			debugNumber = 0; // The first debug has number 0.

    while( !status )
    {
	char 	comment[ 1024 ];
	char*	value;
	char	xdal[ 9 ]; // "XDALnnnn" + null

	sprintf( xdal, "XDAL%d", debugNumber );

	fits_read_key_longstr(
	    fptr,
	    xdal,
	    &value,
	    comment,
	    &status );

	// ------------------------------------
	// Either, there are no more xproc
	// keywords
	// (KEY_NO_EXIST), in which case,
	// do not attempt to read any more
	// and exit from the loop, otherwise,
	// there was a real error.
	// ------------------------------------
	if( status == KEY_NO_EXIST )
	{
	    status = 0;

	    ALTRETURNV( 0 );

	} // if

	fits_check(fptr);

	dataSetServer -> debug( dataSet, value );
	free( value );
	debugNumber++;

    } // while

    // ******************************
    // If we get here an error occured.
    // ******************************

    RETURN 1;

}



int
FitsReaderWriter::writeDebugHistory(
    fitsfile*		fptr,
    const DataSet*		dataSet,
    int&			status ) const
{
    PROC

	if( status != 0 )
	{
	    ALTRETURNV( 1 );
	} // if

    char			xdal[ 9 ]; // "XDALnnnn" + null
    const unsigned int 	debugs = dataSetServer -> debugs( dataSet );

    for( unsigned int i = 0; i < debugs; ++i )
    {
	sprintf( xdal, "XDAL%d", i );
	fits_write_key_longstr(
	    fptr,
	    xdal,
	    (char*)dataSetServer -> debug( dataSet, i ).c_str(),
	    "",
	    &status );
    } // for

    // ******************************
    // Everything is Ok, so return 0.
    // ******************************

    RETURN 0;
}


int
FitsReaderWriter::writeSeparator(
    fitsfile *		fptr,
    unsigned int		nrOfKeywords,
    int &			status ) const
{

    PROC

	if( status != 0 )
	{
	    ALTRETURNV( 1 );
	} // if

    // ************************************************************
    // Only write the separator if there are any keywords to write.
    // ************************************************************
    if( nrOfKeywords > 0 )
    {
	fits_update_key_lng(
	    fptr,
	    (char*)attributeSeparator.c_str(),
	    nrOfKeywords,
	    (char*)( toString( nrOfKeywords ) + attributeSeparatorComment ).c_str(),
	    &status );
    } // if

    // ******************************
    // Everything is Ok, so return 0.
    // ******************************

    RETURN 0;
}

/*
  static string
  removePairedQuotes( string s )
  {
  string		clean;
  bool		expectQuote = false;
  unsigned int	j = 0;

  cout << "s = " << s << endl;

  for( unsigned int i = 0; i < s.length(); ++i )
  {
  if( expectQuote && s[ i ] != '\'' )
  {
  errstr << "There was an error when trying to read a FITS keyword; a quote was expected but not found." << Sas::error( "missingQuoteFromFITSKeyword" );
  }
  else if( expectQuote )
  {
  expectQuote = false;
  // Discard this quote.
  }
  else if( s[ i ] == '\'' )
  {
  // Keep this quote, but will discard the next.
  clean[ j++ ] = s[ i ];
  expectQuote = true;
  }
  else
  {
  clean[ j++ ] = s[ i ];
  } // if
  } // for

  cout << "clean = " << clean << endl;

  return clean;

  }
*/


// **********************************************************
// This function assumes that the single quotes come in pairs
// No checking is done to ensure this.
// **********************************************************
static void
removePairedQuotes( string& s )
{
    string::size_type cp = 0;

    while( (cp = s.find( "''", cp )) != string::npos )
    {
	s.erase( cp++, 1 );
    } // while
}


void
FitsReaderWriter::setAttributeValue(
    Attribute* attribute,
    const string& value) const
{
    PROC

	// FITS string
	if ( value[0] == '\'' )
	{
	    // Strip outer spaces and outer single quotes.
	    string		clean = trim( unquote( value ), true );

	    // ***********************************************************************
	    // It seems that when writing keyword values, CFITSIO writes a
	    // single quote as 2 single quotes.
	    // However, when reading a keyword value, the second quote is not stripped
	    // from the value, giving rise to quotes being duplicated.
	    // It is assumed, therefore that quotes will always be in pairs.
	    // The following removes one of the quotes from each pair.
	    // ***********************************************************************
	    removePairedQuotes( clean );

	    *attribute = trim( unquote( clean ), true );
	} // if

    // FITS boolean
	else if ( value == "F" )
	    *attribute = false;
	else if ( value == "T" )
	    *attribute = true;

    // Floating point
	else if ( value.find_first_of(".EDed") != string::npos ) {
	    stringstream ss;
	    ss << value;
	    double d;
	    ss >> d;
	    if (!ss)
		errstr << "could not interpret the value '" << value << "' of attribute '" << attribute->name() << "' of '" << attribute->parent()->name() << "' as a floating point" << Sas::error("invalidFloatFormat");

	    *attribute = d;
	}

    // Must be integer
	else {
	    stringstream ss;
	    ss << value;
	    int i;
	    ss >> i;
	    if (!ss) errstr << "could not interpret the value '" << value << "' of attribute '" << attribute->name() << "' of '" << attribute->parent()->name() << " as an integer" << Sas::error("invalidFloatFormat");

	    *attribute = i;
	}

    ENDPROC
	}


int
FitsReaderWriter::readLongString(
    fitsfile *		fptr,
    int&			keyNumber,
    char*			value,
    char*			comment,
    int &			status ) const
{
    PROC

	char	v[ 80 ];
    char	c[ 1024 ];
    char	keyName[ 24 ];

    if( status != 0 )
    {
	ALTRETURNV( status );
    } // if

    errstr << "Attempting to read the keyword as a long string" <<
	Sas::message( Msg::Core, Msg::Sparse );

    value[ strlen( value ) - 2 ] = '\'';
    value[ strlen( value ) - 1 ] = '\0';

    fits_read_keyn(
	fptr,
	keyNumber,
	keyName,
	v,
	c,
	&status );

    while( strcmp( keyName, "CONTINUE" ) == 0 )
    {

//  	errstr << "Read continuation keyword at position " << keyNumber
//  	       << " with value '"  << c << "'"
//  	       << Sas::message( Msg::Core, Msg::Verbose ); 

	if( (strlen( c ) == 72) && (c[ 2 ] == '\'') && (c[ 71 ] == '\'') && (c[ 70 ] == '&') )
	{
	    // ============================================
	    // Concatenate the continued string, noting
	    // that fits has prepended c with two spaces.
	    // ============================================
	    value[ strlen( value ) - 1 ] = '\0';
	    strcat( value, c  + 3 );
	    value[ strlen( value ) - 2 ] = '\'';
	    value[ strlen( value ) - 1 ] = '\0';
	}
	else
	{
	    // ==========================================
	    // Split c into value and comment components.
	    // Note that the comment begins with / followed
	    // be a space, which are not copied.
	    // ==========================================
	    const string::size_type	endV = string(c+3).find('\'');
	    const string::size_type	beginC = string(c + 4 + endV).find('/');

	    value[ strlen( value ) - 1 ] = '\0';
	    strncat( value, c + 3, endV + 1 );
	    //AI: Check if a comment has been found within the c string
	    //If not, nothing is copied to comment...
	    if ( beginC != string::npos)
	    	strcpy( comment, c + 6 + endV + beginC );
	}// if

	++keyNumber;

	fits_read_keyn(
	    fptr,
	    keyNumber,
	    keyName,
	    v,
	    c,
	    &status );
	fits_check( fptr );

    } // while

    // ******************************
    // Everything is Ok, so return 0.
    // ******************************


    RETURN 0;
}

int
FitsReaderWriter::readAttributes( 
    fitsfile *		fptr,
    Attributable &		attributeSet,
    KeywordChecker *	attributeChecker,
    int &			status ) const
{

    PROC

	int keyNumber = 1;	// The first keyword in a table
				// has position 1.

    if( status != 0 )
    {
	ALTRETURNV( 1 );
    } // if

    // ************************************************************
    // Read each attribute, in turn, by ordinal position; the first
    // attribute has position 1.
    // ************************************************************
    while( !status )
    {
	char	keyName[ 24 ];
	char	value[ 4*4096 ];
	char	comment[ 1024 ];

	fits_read_keyn(
	    fptr,
	    keyNumber,
	    keyName,
	    value,
	    comment,
	    &status );

	// =======================================================
	// Either, there are no more keywords
	// (KEY_OUT_BOUNDS), in which case,
	// do not attempt to read any more
	// and exit from the loop, otherwise,
	// there was a real error.
	// =======================================================
	// -------------------------------------------------------
	// Strangely sometimes we get no error status but an empty
	// name after the last keyword.
	// -------------------------------------------------------
	// =======================================================
	if( status == KEY_OUT_BOUNDS || strlen( keyName ) == 0 )
	{
	    status = 0;

	    ALTRETURNV( 0 );
	} // if

	//cerr << "read keyword " << keyNumber << " name='" << keyName << "' comment='" << comment << "' value='" << value << "'" << endl;
	SAS_DEBUG( clog << "read keyword " << keyNumber << " name='" << keyName << "' comment='" << comment << "' value='" << value << "'" << endl; )

	    ++keyNumber;

	if( attributeChecker -> ok( keyName ) && !isFITSReservedKeyword( keyName ))
	    // -------------------------
	    // Ignore reserved keywords.
	    // -------------------------
	{
	    // ....................................................
	    // Check to see if the value is a long string.
	    // A Long string designator is a 70-character string
	    // which is delimited by ' and &'
	    // It is assumed that there is no optional comment.
	    // ....................................................
	    if(
		( strlen( value ) == 70 )
		&&
		( value[ 0 ] = '\'' )
		&&
		( value[ 69 ]  == '\'' )
		&&
		( value[ 68 ]  == '&' ))
	    {
		readLongString(
		    fptr,
		    keyNumber,
		    value,
		    comment,
		    status );

	    } // if


	    // ..............
	    // Get the units.
	    // ..............
	    char unitsChar[ 1024 ];

	    fits_read_key_unit(
		fptr,
		keyName,
		unitsChar,
		&status );

	    fits_check(fptr);

	    string units = unitsChar;

	    // Strip units from comment
	    string cleanComment(comment);

	    string::size_type p1 = cleanComment.find( '[' );
	    string::size_type p2 = cleanComment.find( ']' );

	    // Cope with non-standard unit specification at end of comment,
	    // which is not picked up by FITSIO.
	    if ( units.empty() && p1 != string::npos && p2 != string::npos )
		units = cleanComment.substr( p1, p2 );

	    if (p1 != string::npos && p2 != string::npos)
		cleanComment =
		    cleanComment.substr( 0, p1 )
		    + cleanComment.substr( p2 + 1 ,string::npos );
	    cleanComment = trim( cleanComment, true );

	    // *******************************************
	    // Now add the attribute to the attribute set.
	    // *******************************************
	    Attribute	* attribute =

		attributeSet.addAttribute(
		    attributeChecker -> trim( keyName ),
		    cleanComment,
		    units );

	    setAttributeValue( attribute, value );

	    SAS_DEBUG( clog << "read attribute " << *attribute << " from " << fits_file_name( fptr ) << endl; )
		} // if

    } // while

    // **********************************************
    // When we get here we have encountered an error.
    // **********************************************

    RETURN 1;

}


int
FitsReaderWriter::readCommentsAndHistory(
    fitsfile *		fptr,
    Attributable &		attributeSet,
    KeywordChecker *	attributeChecker,
    int &			status ) const
{
    PROC

	if( status != 0 )
	{
	    ALTRETURNV( 1 );
	} // if

    int			keyNumber;
    char			keyName[ 1024 ];
    char 			comment[ 1024 ];
    char			value[ 1024 ];

    keyNumber = 1;		// The first keyword in a table
				// has position 1.

    while( !status )
    {
	fits_read_keyn(
	    fptr,
	    keyNumber,
	    keyName,
	    value,
	    comment,
	    &status );

	// ------------------------------------
	// Either, there are no more keywords
	// (KEY_OUT_BOUNDS), in which case,
	// do not attempt to read any more
	// and exit from the loop, otherwise,
	// there was a real error.
	// ------------------------------------
	if (status == KEY_OUT_BOUNDS)
	{

	    status = 0;

	    ALTRETURNV( 0 );

	} // if

	fits_check(fptr);

	if( attributeChecker -> ok( keyName ) )
	{
	    // *********************************************
	    // The required comment/history text has been
	    // written to the char pointer variable
	    // 'comment'. For some reason, the comment/history
	    // is prepended with two spaces (if it is not
	    // null), which have to
	    // be removed (otherwise, when the comment
	    // is written out again to the FITS file,
	    // the two spaces have the effect of
	    // indenting the comment/history).
	    // Removal of the two spaces is acheived by
	    // adding 2 to the character pointer.
	    // *********************************************
	    if( strcmp( keyName, "COMMENT" ) == 0 )
	    {
		if( strlen( comment ) == 0 )
		{
		    errstr << "A blank comment (for reading) has been detected in the attributable set with qualified name '" << attributeSet.qualifiedName() << "'.  Since FITS does not handle these correctly, the null string will be used." << Sas::message( Msg::Core, Msg::Verbose );

		    attributeSet.addComment( comment );
		}
		else
		{
		    attributeSet.addComment(
			comment
			+ numberOfFITSSpaces );
		} // if

	    }
	    else if( strcmp( keyName, "HISTORY" ) == 0 )
	    {
		if( strlen( comment ) == 0 )
		{
		    errstr << "A blank history record (for reading) has been detected in the attributable set with qualified name '" << attributeSet.qualifiedName() << "'.  Since FITS does not handle these correctly, the null string will be used." << Sas::message( Msg::Core, Msg::Verbose );
		}
		else
		{
		    attributeSet.addHistory(
			comment
			+ numberOfFITSSpaces );
		} // if
	    } // if
	} // if
	keyNumber++;

    } // while

    // ******************************
    // If we get here an error occured.
    // ******************************

    RETURN 1;

}


int
FitsReaderWriter::readHeader(
    fitsfile *		fptr,
    Attributable &		attributeSet,
    KeywordChecker *	attributeChecker,
    int &			status ) const
{

    PROC

	if( status != 0 )
	{
	    ALTRETURNV( 1 );
	} // if

    readAttributes( fptr, attributeSet, attributeChecker, status );
    attributeChecker -> reset();
    readCommentsAndHistory( fptr, attributeSet, attributeChecker, status );

    // ******************************
    // Everything is Ok, so return 0.
    // ******************************

    RETURN 0;

}


int
FitsReaderWriter::writeColumnScalingFactors(
    fitsfile *	fptr,
    const Column *	column,
    int		columnNumber,
    int &		status ) const
{
    PROC

	if( status != 0 )
	{
	    ALTRETURNV( 1 );
	} // if

    if( column -> scaled() )
    {
	double		value;

	{
	    stringstream ss;
	    ss << "TZERO" << columnNumber + 1;
	    string format=ss.str();
	    value = column -> scaleZero();
	    fits_update_key_dbl(
		fptr,
		(char *)format.c_str(),
		value,
		0,
		0,
		&status );
	}

	{
	    stringstream ss;
	    ss << "TSCAL" << columnNumber  + 1;
	    string format=ss.str();
	    value = column -> scaleFactor();

	    fits_update_key_dbl(
		fptr,
		(char *)format.c_str(),
		value,
		0,
		0,
		&status );
	}

	// ***********************************
	// Disable the FITS automatic scaling.
	// ***********************************
	fits_set_tscale(
	    fptr,
	    columnNumber + 1,
	    1.0,
	    0.0,
	    &status );

	fits_check(fptr);

    } // if

    // ******************************
    // Everything is Ok, so return 0.
    // ******************************

    RETURN 0;

}


int
FitsReaderWriter::writeArrayNullValue(
    fitsfile *	fptr,
    const Array *	array,
    int &		status ) const
{
    PROC

	if( status != 0 )
	{
	    ALTRETURNV( 1 );
	} // if

    if( array -> nullType() == Nullable::Integer )
    {
	string format("BLANK");
	int		value;
	value = array -> intNullValue();

	fits_update_key_lng(
	    fptr,
	    (char *)format.c_str(),
	    value,
	    (char*)arrayNullValueComment.c_str(),
	    &status );

	fits_check(fptr);

    } // if

    // ******************************
    // Everything is Ok, so return 0.
    // ******************************

    RETURN 0;

}



int
FitsReaderWriter::writeColumnNullValue(
    fitsfile *	fptr,
    const Column *	column,
    int &		status ) const
{
    PROC

	if( status != 0 )
	{
	    ALTRETURNV( 1 );
	} // if

    if( column -> nullType() == Nullable::Integer )
    {
	stringstream  ss;
	int		value;

	ss << "TNULL" << column -> columnNumber() + 1;
	value = column -> intNullValue();

	string format=ss.str();
	fits_update_key_lng(
	    fptr,
	    (char *)format.c_str(),
	    value,
	    (char*)columnNullValueComment.c_str(),
	    &status );

	fits_check(fptr);

    } // if

    // ******************************
    // Everything is Ok, so return 0.
    // ******************************

    RETURN 0;

}


int
FitsReaderWriter::writeColumnComment(
    fitsfile *	fptr,
    const string &	comment,
    unsigned int	columnNr,
    int &		status ) const
{
    if( status != 0 )
    {
	ALTRETURNV( 1 );
    } // if

    // **************************
    // Construct the name TTYPEn.
    // **************************
    ostringstream ss;
	
    ss << "TTYPE" << columnNr + 1;
    string ttype=ss.str();

    // *******************************************
    // Now modify the comment field of the column.
    // *******************************************
    fits_modify_comment(
	fptr,
	(char*)ttype.c_str(),
	comment == "" ? (char*)defaultColumnComment.c_str() : (char*)comment.c_str(),
	&status );

    fits_check(fptr);

    // ******************************
    // Everything is Ok, so return 0.
    // ******************************

    RETURN 0;

}


bool
FitsReaderWriter::columnAttributeOK( const string & name ) const
{
    PROC

	bool	ok = true;

    // *****************************************************************
    // Column Attributes must be one of the allowable column attributes.
    // *****************************************************************
    if( !isFITSColumnKeyword( name ))
    {
	errstr << "The attribute with name '" << name << "' is not legal for a column attribute since it is not one of the reserved attribute names.  It is therefore being ignored and will not be written to the Fits file." << Sas::warning( "IllegalColumnAttributeName" );
	ok = false;
    } // if

    RETURN ok;
}


bool
FitsReaderWriter::tableAttributeOK( const string & attributeName ) const
{
    PROC

	bool	ok = true;

    // ***********************************************
    // Table Attributes must not be a reserved
    // column attribute ending with an integer.
    // ***********************************************
    if( isFITSColumnKeyword( trimDigits( attributeName )))
    {
	errstr << "The attribute with name '" << attributeName << "' is not legal since it is one of the reserved column attributes ending with a digit.  It is therefore being ignored and will not be written to the Fits file." << Sas::warning( "IllegalTableAttributeName" );
	ok = false;
    } // if

    RETURN ok;
}


int
FitsReaderWriter::writeArrayUnits(
    fitsfile *	fptr,
    const Array &	array,
    int &		status ) const
{
    PROC

	if( status != 0 )
	{
	    ALTRETURNV( 1 );
	} // if

    if( array.units() != "" )
    {
	fits_update_key_str(
	    fptr,
	    "BUNIT",	// This is how to set the array units.
	    (char*)array.units().c_str(),
	    (char*)arrayUnitsComment.c_str(),
	    &status );

    } // if

    // ******************************
    // Everything is Ok, so return 0.
    // ******************************
    RETURN 0;
}


int
FitsReaderWriter::checkAndSetUnsignedArray(
    fitsfile *		fptr,
    Array::DataType&	dataType,
    int &			status ) const
{
    PROC

	if( status != 0 )
	{
	    ALTRETURNV( 1 );
	} // if

    // ******************************************
    // We are only interested in int16 and int32.
    // ******************************************
    if(
	dataType != Array::Int16
	&&
	dataType != Array::Int32 )
    {
	// *****************
	// Everything is Ok.
	// *****************

	ALTRETURNV( 0 );
    } // if

    char 		comment[ 1024 ];
    double		scale = 0.0;
    double		zero = 0.0;
    string	zerostr("BZERO");
    string	scalstr("BSCALE");

    // ********************************************
    // Move pointer to beginning of current header.
    // ********************************************
    fits_read_record(
	fptr,
	0,
	comment,
	&status );

    // ********************
    // Look for zero/scale.
    // ********************
    fits_read_key_dbl(
	fptr,
	(char*)zerostr.c_str(),
	&zero,
	comment,
	&status );

    fits_read_key_dbl(
	fptr,
	(char*)scalstr.c_str(),
	&scale,
	comment,
	&status );

    if( status == KEY_NO_EXIST )
    {
	status = 0;

	// *****************
	// Everything is Ok.
	// *****************

	ALTRETURNV( 0 );
    } // if

    fits_check(fptr);

    // *****************************************************
    // We now know that status is zero, and consequently
    // the scaling/zero factors were found.
    // Check that they are consistent with unsigned scaling,
    // and set dataType to unsigned as appropriate.
    // *****************************************************
    if(
	dataType == Array::Int16
	&&
	zero == 32768.0
	&&
	scale == 1.0 )
    {	
	dataType = Array::Uint16;
    }
    else if(
	dataType == Array::Int32
	&&
	zero == 2147483648.0
	&&
	scale == 1.0 )
    {
	dataType = Array::Uint32;
    } // if

    // *****************
    // Everything is Ok.
    // *****************

    RETURN 0;

}


/*
  int
  FitsReaderWriter::checkAndSetUnsignedColumn(
  fitsfile *		fptr,
  int			columnNumber,
  Column::DataType&	dataType,
  int &			status ) const
  {
  PROC

  if( status != 0 )
  {
  ALTRETURNV( 1 );
  } // if

  // ******************************************
  // We are only interested in int16 and int32.
  // ******************************************
  if(
  dataType != Column::Int16
  &&
  dataType != Column::Int32 )
  {
  // *****************
  // Everything is Ok.
  // *****************

  ALTRETURNV( 0 );
  } // if

  char 		comment[ 1024 ];
  double		scale = 0.0;
  double		zero = 0.0;
  stringstream	zerostr;
  stringstream	scalstr;
	
  zerostr << "TZERO" << columnNumber + 1;
  scalstr << "TSCAL" << columnNumber + 1;

  // ********************************************
  // Move pointer to beginning of current header.
  // ********************************************
  fits_read_record(
  fptr,
  0,
  comment,
  &status );

  // ********************
  // Look for zero/scale.
  // ********************
  fits_read_key_dbl(
  fptr,
  zerostr.str().c_str(), // JBa TODO: FIX DANGEROUS
  &zero,
  comment,
  &status );

  fits_read_key_dbl(
  fptr,
  scalstr.str().c_str(), // JBa TODO: FIX DANGEROUS
  &scale,
  comment,
  &status );

  if( status == KEY_NO_EXIST )
  {
  status = 0;

  // *****************
  // Everything is Ok.
  // *****************

  ALTRETURNV( 0 );
  } // if

  fits_check(fptr);

  // *****************************************************
  // We now know that status is zero, and consequently
  // the scaling/zero factors were found.
  // Check that they are consistent with unsigned scaling,
  // and set dataType to unsigned as appropriate.
  // *****************************************************
  if(
  dataType == Column::Int16
  &&
  zero == 32768.0
  &&
  scale == 1.0 )
  {	
  dataType = Column::Uint16;
  }
  else if(
  dataType == Column::Int32
  &&
  zero == 2147483648.0
  &&
  scale == 1.0 )
  {
  dataType = Column::Uint32;
  } // if

  // *****************
  // Everything is Ok.
  // *****************

  RETURN 0;

  }
*/

/*AI gcc 4 porting 
Line 1723: suggest parentheses around && within ||
*/
void
FitsReaderWriter::checkAndSetUnsignedColumn(
    ColumnInfo&	columnInfo ) const
{
    PROC

	// ******************************************
	// We are only interested in int16 and int32.
	// ******************************************
	if(
	    columnInfo._dataType != Column::Int16
	    &&
	    ( columnInfo._dataType != Column::Int32
	    ||
	      !columnInfo._scalingFactorsFound ) )
	{
	    ALTRETURN
		} // if

    // *****************************************************
    // The scaling/zero factors were found.
    // Check that they are consistent with unsigned scaling,
    // and set dataType to unsigned as appropriate.
    // *****************************************************
    if(
	columnInfo._dataType == Column::Int16
	&&
	columnInfo._zero == 32768.0
	&&
	columnInfo._scale == 1.0 )
    {	
	columnInfo._dataType = Column::Uint16;
    }
    else if(
	columnInfo._dataType == Column::Int32
	&&
	columnInfo._zero == 2147483648.0
	&&
	columnInfo._scale == 1.0 )
    {
	columnInfo._dataType = Column::Uint32;
    } // if
    ENDPROC
	}


int
FitsReaderWriter::getArrayDataType(
    fitsfile *		fptr,
    int&			bitpix,
    Array::DataType&	dataType,
    int &			status ) const
{
    PROC

	if( status != 0 )
	{
	    ALTRETURNV( 1 );
	} // if

    // ************************
    // Get the array data type.
    // ************************
    fits_read_key(
	fptr,
	TINT,
	"BITPIX",
	&bitpix,
	0,
	&status );

    fits_check(fptr);

    dataType = fromFITSBitpix( bitpix );
	
    // *****************
    // Everything is Ok.
    // *****************

    RETURN 0;

}


int
FitsReaderWriter::getColumnDataType(
    fitsfile *		fptr,
    int			columnNumber,
    int&			typeCode,
    Column::DataType&	dataType,
    int &			status ) const
{
    PROC

	if( status != 0 )
	{
	    ALTRETURNV( 1 );
	} // if

    long			repeat;
    long			width;

    // *************************
    // Get the column data type.
    // *************************
    fits_get_coltype(
	fptr,
	columnNumber + 1,
	&typeCode,
	&repeat,
	&width,
	&status );
    fits_check(fptr);

    if( isValidDalDataType( abs( typeCode )) )
    {
	try {
	    SetExceptionErrorHandler eeh;
	    dataType = fromFITSDataType( abs( typeCode ));
	}	
	catch(SasException::Error e){
	    char n[256];
	    fits_file_name(fptr, n, &status);
	    errstr << "There was an error while trying to determine the type of column number " << columnNumber << " in the FITS file " << std::string(n)
		   << ". The error said: [" << e.code << "] "
		   << e.message << Sas::error("FITSError");
	}
    } // if

    // *****************
    // Everything is Ok.
    // *****************

    RETURN 0;

}


int
FitsReaderWriter::readColumnNullValue(
    fitsfile *	fptr,
    Column &	column,
    int &		status ) const
{
    PROC

	if( status != 0 )
	{
	    ALTRETURNV( 1 );
	} // if

    char 		comment[ 1024 ];
    long		nullValue;
    stringstream  ss;

    ss << "TNULL" << column.columnNumber() + 1;
    string nullstr=ss.str();

    // ********************************************
    // Move pointer to beginning of current header.
    // ********************************************
    fits_read_record(
	fptr,
	0,
	comment,
	&status );

    // ********************
    // Look for null value.
    // ********************
    fits_read_key_lng(
	fptr,
	(char*)nullstr.c_str(), // JBa TODO: FIX DANGEROUS
	&nullValue,
	comment,
	&status );

    if( status == 0 )
    {
	column.nullValue( nullValue );
    } // if

    if (status == KEY_NO_EXIST) status = 0;
    fits_check(fptr);

    // *****************
    // Everything is Ok.
    // *****************

    RETURN 0;

}


int
FitsReaderWriter::readArrayNullValue(
    fitsfile *	fptr,
    Array&		array,
    int &		status ) const
{
    PROC

	if( status != 0 )
	{
	    ALTRETURNV( 1 );
	} // if

    char 		comment[ 1024 ];
    long		nullValue;
    string nullstr("BLANK");

    // ********************************************
    // Move pointer to beginning of current header.
    // ********************************************
    fits_read_record(
	fptr,
	0,
	comment,
	&status );

    // ********************
    // Look for null value.
    // ********************
    fits_read_key_lng(
	fptr,
	(char*)nullstr.c_str(), // JBa TODO: FIX DANGEROUS
	&nullValue,
	comment,
	&status );

    if( status == 0 )
    {
	array.nullValue( nullValue );
    } // if

    if (status == KEY_NO_EXIST) status = 0;
    fits_check(fptr);

    // *****************
    // Everything is Ok.
    // *****************

    RETURN 0;

}


int
FitsReaderWriter::readArrayScalingFactors(
    fitsfile	* fptr,
    Array &		array,
    int &		status ) const
{
    PROC

	if( status != 0 )
	{
	    ALTRETURNV( 1 );
	} // if

    char 		comment[ 1024 ];
    double		scale;
    double		zero;

    // ********************************************
    // Move pointer to beginning of current header.
    // ********************************************
    fits_read_record(
	fptr,
	0,
	comment,
	&status );

    // **************
    // Look for Zero.
    // **************
    fits_read_key_dbl(
	fptr,
	"BZERO",
	&zero,
	comment,
	&status );

    // ***************
    // Look for Scale.
    // ***************
    fits_read_key_dbl(
	fptr,
	"SCALE",
	&scale,
	comment,
	&status );

    if( status == 0 )
    {

	// ***************
	// Look for Scale.
	// ***************
	fits_read_key_dbl(
	    fptr,
	    "SCALE",
	    &scale,
	    comment,
	    &status );

	// **********************************************************
	// If the scaling factors where found, set the array scaling,
	// and disable the FITS automatic scaling.
	// **********************************************************
	if( status == 0 )
	{

	    array.scaling( zero, scale );

	    // ***********************************
	    // Disable the FITS automatic scaling.
	    // ***********************************
	    fits_set_bscale(
		fptr,
		1.0,
		0.0,
		&status );

	    fits_check(fptr);
	} // if
    } // if

    if (status == KEY_NO_EXIST) status = 0;
    fits_check(fptr);

    // *****************
    // Everything is Ok.
    // *****************

    RETURN 0;

}


int
FitsReaderWriter::getBlockNameAndComment(
    fitsfile*		fptr,
    const unsigned int&	blockNumber,
    string&			name,
    string&			comment,
    int&			status ) const
{
    PROC

	char			blockName[ 1024 ];
    char 			com[ 1024 ];

    if( status != 0 )
    {
	ALTRETURNV( 1 );
    } // if


    // ===================================================
    // Get the blockname. This is achieved by getting the
    // FITS file keyword EXTNAME.
    // ===================================================
    fits_read_keyword(
	fptr,
	"EXTNAME",	// This is how to get the name.
	blockName,
	com,
	&status );

    switch (status) {

	case 0:

	    name = trim(unquote(blockName));
	    comment = com;

	    break;

	case KEY_NO_EXIST:

	    // --------------------------------------------
	    // The block name keyword EXTNAME was not
	    // found. Construct a unique name for the
	    // block. This is not a fool-proof
	    // way of doing it, but it should
	    // be alright most of the time.
	    // *******************************
	    name = "NONAME" + toString( blockNumber );
	    comment = "";
	    status = 0;		

	    break;

	default :

	    fits_check(fptr);

	    break;

    } // switch

    SAS_DEBUG( clog << "encountered block " << blockName << " in " << fits_file_name( fptr ) << endl; )

	// ******************************
	// Everything is Ok, so return 0.
	// ******************************
	RETURN 0;

}


int
FitsReaderWriter::readColumnUnits(
    fitsfile *	fptr,
    int 		columnNumber,
    char *		units,
    char *		comment,
    int &		status ) const
{
    PROC

	if( status != 0 )
	{
	    ALTRETURNV( 1 );
	} // if

    fits_read_key(
	fptr,
	TSTRING,
	(char *)( "TUNIT" + toString( columnNumber + 1 )).c_str(),
	units,
	comment,
	&status );

    // ------------------------------------
    // Either, no column units are defined
    // (KEY_NO_EXIST), in which case,
    // null terminate the units string
    // and exit, otherwise,
    // there was a real error.
    // ------------------------------------
    if (status == KEY_NO_EXIST) {

	units[0] = 0;
	status = 0;
    }

    fits_check(fptr);

    RETURN 0;

}


int
FitsReaderWriter::readArrayDimensions(
    fitsfile *	fptr,
    int &		dimensions,
    long *		size,
    int &		status ) const
{

    PROC

	string		naxis = "NAXIS";
    char		comment[ 1024 ];


    if( status != 0 )
    {
	ALTRETURNV( 1 );
    } // if


    // ******************************
    // read the number of dimensions.
    // ******************************
    fits_read_key(
	fptr,
	TINT,
	(char *)naxis.c_str(),
	&dimensions,
	comment,
	&status );

    fits_check(fptr);

    // *************************
    // Get each dimension size.
    // *************************
    for( int i = 0; i < dimensions; i++ )
    {
	// ********************************************************
	// read the dimension size.
	// The indices are reverted since FITS uses
	// the Fortran convention (NAXIS1 runs faster than NAXIS2).
	// ********************************************************
	fits_read_key_lng(
	    fptr,
	    (char *)(naxis + toString( i + 1 )).c_str(),
	    &size[dimensions - 1 - i],
	    comment,
	    &status );

	fits_check(fptr);
    } // for

    SAS_DEBUG( clog << "size="; for (int i = 0; i < dimensions; ++i) clog << ' ' << size[i]; clog << endl; )

	// ******************************
	// Everything is Ok, so return 0.
	// ******************************
	RETURN 0;


}


int
FitsReaderWriter::readArrayUnits(
    fitsfile *	fptr,
    string &	units,
    int &		status ) const
{
    PROC

	char		comment[ 1024 ];
    char		unit[ 1024 ];

    if( status != 0 )
    {
	ALTRETURNV( 1 );
    } // if

    // ********************************************
    // Move pointer to beginning of current header.
    // ********************************************
    fits_read_record(
	fptr,
	0,
	comment,
	&status );

    // ***********************************
    // Look for Array Units keyword BUNIT.
    // ***********************************
    fits_read_key(
	fptr,
	TSTRING,
	"BUNIT",
	&unit,
	comment,
	&status );

    if (status && status != KEY_NO_EXIST) fits_check(fptr);

    // *********************************************************
    // If the BUNIT keyword was found then set the units to the
    // found value, else set the units to a null string.
    // *********************************************************
    if( status == 0 )
    {
	units = unit;
    }
    else
    {
	units = "";
    } // if

    // ******************************
    // Everything is Ok, so return 0.
    // ******************************

    status = 0;

    RETURN 0;

}


int
FitsReaderWriter::writeLongStringWarning(
    fitsfile *	fptr,
    int &		status ) const
{
    if( status != 0 )
    {
	ALTRETURNV( 1 );
    } // if

    // **********************
    // Move to primary array.
    // **********************
    int		hdutype;
    int		hdunum = 1;		// Primary header in position 1.

    fits_movabs_hdu( fptr, hdunum, &hdutype, &status );
    fits_check( fptr );


    // *************************************************************
    // TEMPORILY DISABLE because the generated comments are getting
    // confused with other general comment.
    // Write the LONGSTRN warning message to the FITS file
    // to indicate that the FITS long string (string-keyword values
    // greater than 68) is being used. This does not indicate that
    // there are any keywords with values longer than 68 characters,
    // only that there may be.
    // *************************************************************
    //fits_write_key_longwarn( fptr, &status );
    //fits_check( fptr );


    // ******************************
    // Everything is Ok, so return 0.
    // ******************************

    RETURN 0;

}


bool
FitsReaderWriter::acceptTableKeyword(
    fitsfile*	fptr,
    const string&	keyName ) const
{
    PROC
	if( isFITSReservedKeyword( keyName ) )
	{
	    errstr << "The reserved keyword '" << keyName << "' was found in a table extension in the fits file with name '" << fits_file_name( fptr ) << "', but is not allowed in this context. It'll be ignored." << Sas::message( Msg::Core,Msg::Sparse);
	    return true;
	} // if
    RETURN false;
}


int
FitsReaderWriter::readTableInfo(
    fitsfile*	fptr,
    //const DataSet&	dataSet,
    TableInfo&	tableInfo,
    int&		status ) const
{
    if( status != 0 )
    {
	ALTRETURNV( 1 );
    } // if

    // ************************************************************
    // Read each attribute, in turn, by ordinal position; the first
    // attribute has position 1.
    // ************************************************************
    const unsigned int	maxdim = 4;
    int			naxis;
    long			naxes[ maxdim ];
    char			keyName[ 24 ];
    char			value[ 4*4096 ];
    char			comment[ 1024 ];
    int			keyNumber = 1;	// The first keyword in a table has position 1.
    unsigned int		columnNumber;
    int			typeCode;
    long			repeat;
    long			width;
    //const unsigned int	blockNumber( dataSet.blocks() );

    while( !status )
    {
	fits_read_keyn(
	    fptr,
	    keyNumber,
	    keyName,
	    value,
	    comment,
	    &status );

	// =======================================================
	// Either, there are no more keywords
	// (KEY_OUT_BOUNDS), in which case,
	// do not attempt to read any more
	// and exit from the loop, otherwise,
	// there was a real error.
	// =======================================================
	// -------------------------------------------------------
	// Strangely sometimes we get no error status but an empty
	// name after the last keyword.
	// -------------------------------------------------------
	// =======================================================
	if( status == KEY_OUT_BOUNDS || strlen( keyName ) == 0 )
	{
	    status = 0;

	    break;
	} // if

	fits_check();

	errstr << "Found keyword '" <<
	    keyName << "' at position " << keyNumber <<
	    " comment = '" << comment << "' value = '" << value <<
	    "' strlen(value) = " << strlen( value ) <<
	    Sas::message( Msg::Core, Msg::Verbose ); 

	++keyNumber;

	// ***********************************
	// Check for extension name and label.
	// ***********************************
	if( strcmp( keyName, "EXTNAME" ) == 0 )
	{
	    tableInfo._name = trim( unquote( value ));
	    tableInfo._label = comment;
	}
	// **************************************************
	// Check for NAXIS2 (the number of rows in the table.
	// **************************************************
	else if( strcmp( keyName, "NAXIS2" ) == 0 )
	{
	    tableInfo._rows = atoi( value );
	}
	// *************************
	// Check for comment record.
	// *************************
	else if( strcmp( keyName, "COMMENT" ) == 0 )
	{
	    if( strlen( comment ) == 0 )
	    {
		errstr << "A blank comment (for reading) has been detected.  Since FITS does not handle these correctly, the null string will be used." << Sas::message( Msg::Core, Msg::Verbose );

		tableInfo._comments.push_back( comment );
	    }
	    else
	    {
		tableInfo._comments.push_back(
		    comment
		    + numberOfFITSSpaces );
	    } // if
	}
	// *************************
	// Check for history record.
	// *************************
	else if( strcmp( keyName, "HISTORY" ) == 0 )
	{
	    if( strlen( comment ) == 0 )
	    {
		errstr << "A blank history record (for reading) has been detected.  Since FITS does not handle these correctly, the null string will be used." << Sas::message( Msg::Core, Msg::Verbose );

		tableInfo._history.push_back( comment );
	    }
	    else
	    {
		tableInfo._history.push_back(
		    comment
		    + numberOfFITSSpaces );
	    } // if
	}
	// **************************
	// Check for column datatype.
	// **************************
	else if( strncmp( keyName, "TFORM", 5 ) == 0 )
	{
	    sscanf( keyName + 5, "%u", &columnNumber );
	    const string	cleanValue( trim( unquote( value )));

	    fits_binary_tform( (char*)(cleanValue.c_str()), &typeCode, &repeat, &width, &status );
	    fits_check();

	    tableInfo._columns[ columnNumber - 1 ]._typeCode = typeCode;

	    longVec	size;

	    if( repeat > 1 )
	    {
		size.push_back( repeat );
	    }
	    else if( typeCode > 0 )
	    {
		size = Dal::scalar;
	    }
	    else
	    {
		size = Dal::zero;
	    }// if

	    tableInfo._columns[ columnNumber - 1 ]._size = size;
	    tableInfo._columns[ columnNumber - 1 ]._width = width;
	    tableInfo._columns[ columnNumber - 1 ]._repeat = repeat;
	}
	// ****************************
	// Check for column dimemsions.
	// ****************************
	else if( strncmp( keyName, "TDIM", 4 ) == 0 )
	{
	    sscanf( keyName + 4, "%u", &columnNumber );
	    fits_decode_tdim(
		fptr,
		value,
		columnNumber,
		maxdim,
		&naxis,
		naxes,
		&status );
	    fits_check();
	    tableInfo._columns[ columnNumber - 1 ]._size = vector<unsigned long>(
		(unsigned long *)(naxes),
		(unsigned long *)(naxes + naxis) );

	}
	// ********************************
	// Check for column name and label. 
	// ********************************
	else if( strncmp( keyName, "TTYPE", 5 ) == 0 )
	{
	    sscanf( keyName + 5, "%u", &columnNumber );
	    tableInfo._columns[ columnNumber - 1 ]._name = trim( unquote( value ));
	    tableInfo._columns[ columnNumber - 1 ]._label = comment;
	}
	// ***********************
	// Check for column units.
	// ***********************
	else if( strncmp( keyName, "TUNIT", 5 ) == 0 )
	{
	    sscanf( keyName + 5, "%u", &columnNumber );
	    tableInfo._columns[ columnNumber - 1 ]._units = trim( unquote( value ));
	}
	// ****************************
	// Check for column null value.
	// ****************************
	else if( strncmp( keyName, "TNULL", 5 ) == 0 )
	{
	    sscanf( keyName + 5, "%u", &columnNumber );
	    tableInfo._columns[ columnNumber - 1 ]._nullValue = atoi( value );
	    tableInfo._columns[ columnNumber - 1 ]._nullValueFound = true;
	}
	// ***********************
	// Check for column TZERO.
	// ***********************
	else if( strncmp( keyName, "TZERO", 5 ) == 0 )
	{
	    sscanf( keyName + 5, "%u", &columnNumber );
	    tableInfo._columns[ columnNumber - 1 ]._zero = atof( value );
	    tableInfo._columns[ columnNumber - 1 ]._scalingFactorsFound = true;
	}
	// ***********************
	// Check for column TSCAL.
	// ***********************
	else if( strncmp( keyName, "TSCAL", 5 ) == 0 )
	{
	    sscanf( keyName + 5, "%u", &columnNumber );
	    tableInfo._columns[ columnNumber - 1 ]._scale = atof( value );
	    tableInfo._columns[ columnNumber - 1 ]._scalingFactorsFound = true;
	}
	// ***************************
	// Check for column attribute.
	// ***************************
	else if( isFITSColumnKeyword( trimDigits( keyName )))
	{
	    sscanf( keyName + 5, "%u", &columnNumber );

	    // *************************************************
	    // Check to see if the value is a long string.
	    // A Long string designator is a 70-character string
	    // which is delimited by ' and &'
	    // It is assumed that there is no optional comment.
	    // *************************************************
	    if(
		( strlen( value ) == 70 )
		&&
		( value[ 0 ] = '\'' )
		&&
		( value[ 69 ]  == '\'' )
		&&
		( value[ 68 ]  == '&' ))
	    {
		readLongString(
		    fptr,
		    keyNumber,
		    value,
		    comment,
		    status );

	    } // if

	    // *********************************
	    // Get and Strip units from comment.
	    // *********************************
	    string			cleanComment( comment );
	    string			units;
	    string::size_type	p1( cleanComment.find( '[' ) );
	    string::size_type	p2( cleanComment.find( ']' ) );

	    // ============================================================
	    // Cope with non-standard unit specification at end of comment,
	    // which is not picked up by FITSIO.
	    // ============================================================
	    if( p1 != string::npos && p2 != string::npos )
	    {
		units = cleanComment.substr( p1 + 1, p2 - p1 - 1 );
	    } // if

	    if( p1 != string::npos && p2 != string::npos )
	    {
		cleanComment =
		    cleanComment.substr( 0, p1 )
		    + cleanComment.substr( p2 + 1, string::npos );
	    } // if
	    cleanComment = trim( cleanComment, true );

	    AttributeInfo	ai;
	    ai._name = trimDigits( keyName );
	    ai._value = value;
	    ai._label = cleanComment;
	    ai._units = units;
	    tableInfo._columns[ columnNumber - 1 ]._attributes.push_back( ai );
	}
	else if( strcmp( keyName, "TFIELDS" ) == 0 )
	{
	    tableInfo._numberOfColumns = atoi( value );
	}
	// ****************************
	// Skip the following keywords.
	// ****************************
	else if(
	    strcmp( keyName, "XTENSION" ) == 0 
	    ||
	    strcmp( keyName, "BITPIX" ) == 0
	    ||
	    strcmp( keyName, "NAXIS" ) == 0
	    ||
	    strcmp( keyName, "NAXIS1" ) == 0
	    ||
	    strcmp( keyName, "PCOUNT" ) == 0
	    ||
	    strcmp( keyName, "GCOUNT" ) == 0
	    ||
	    strcmp( keyName, "END" ) == 0 )
	{
	}
	// ******************************************************
	// Check that all table reserved words have been handled.
	// ******************************************************
	else if( acceptTableKeyword( fptr, keyName ) )
	{
	    // Nothing to do
	}
	// **********************************************
	// Assume that what is left is a table attribute.
	// **********************************************
	else
	{
	    // *************************************************
	    // Check to see if the value is a long string.
	    // A Long string designator is a 70-character string
	    // which is delimited by ' and &'
	    // It is assumed that there is no optional comment.
	    // *************************************************
	    if(
		( strlen( value ) == 70 )
		&&
		( value[ 0 ] = '\'' )
		&&
		( value[ 69 ]  == '\'' )
		&&
		( value[ 68 ]  == '&' ))
	    {
		readLongString(
		    fptr,
		    keyNumber,
		    value,
		    comment,
		    status );

		errstr <<"Long string keyword '" << keyName
		       << "' comment = '"
		       << comment << "' value = '" << value
		       << "' strlen(value) = " << strlen( value )
		       << Sas::message( Msg::Core, Msg::Verbose ); 

	    } // if

	    // *********************************
	    // Get and Strip units from comment.
	    // *********************************
	    string			cleanComment( comment );
	    string			units;
	    string::size_type	p1( cleanComment.find( '[' ) );
	    string::size_type	p2( cleanComment.find( ']' ) );

	    // ============================================================
	    // Cope with non-standard unit specification at end of comment,
	    // which is not picked up by FITSIO.
	    // ============================================================
	    if( p1 != string::npos && p2 != string::npos )
	    {
		units = cleanComment.substr( p1 + 1, p2 - p1 - 1 );
	    } // if

	    if( p1 != string::npos && p2 != string::npos )
	    {
		cleanComment =
		    cleanComment.substr( 0, p1 )
		    + cleanComment.substr( p2 + 1, string::npos );
	    } // if
	    cleanComment = trim( cleanComment, true );

	    AttributeInfo	ai;
	    ai._name = keyName;
	    ai._value = value;
	    ai._label = cleanComment;
	    ai._units = units;
	    tableInfo._attributes.push_back( ai );

	} // if
    } // while

    // ***************************
    // Do some consistancy checks.
    // ***************************
    // ......................
    // ........ TODO .........
    // ......................

    if(tableInfo._numberOfColumns != tableInfo._columns.size() ){
	errstr << "The number of columns reported by the TFIELDS attribute is "
	       << tableInfo._numberOfColumns
	       << " whereas I have detected " << tableInfo._columns.size()
	       << " data columns. Make sure the data set is consistent."
	       << Sas::error("InconsistentDataSet");
    }

    // ************************************
    // Number the columns and fits pointer.
    // ************************************
    for( unsigned int i = 0; i < tableInfo._columns.size(); ++i )
    {
	tableInfo._columns[ i ]._columnNumber = i;
	tableInfo._columns[ i ]._fptr = fptr;
    } // for

    // ******************************
    // Everything is Ok, so return 0.
    // ******************************
    RETURN 0;

}


Table*
FitsReaderWriter::addTable(
    DataSet&		dataSet,
    TableInfo&		tableInfo ) const
{
    PROC
	const unsigned int	blockNumber( dataSet.blocks() );

    // ******************************
    // Add the table to the data set.
    // ******************************
    if( tableInfo._name.empty() )
    {
	tableInfo._name = "NONAME" + toString( blockNumber );
	errstr << "I found a table without a name. It was renamed: "
	       << tableInfo._name << Sas::warning( "unnamedTable" );

    } // if

    // *****************************************************
    // If the name already existed, append the block number.
    // *****************************************************
    if ( dataSet.hasBlock( tableInfo._name ) )
    {
	errstr << "I found a table called " << tableInfo._name
	       << " but there exists another table with such a name. I have renamed the table just found ";
	tableInfo._name += toString( blockNumber );
	errstr << tableInfo._name << Sas::warning( "duplicateTableName" );	
    } // if

    Table *		table = dataSet.addTable(
	tableInfo._name,
	tableInfo._rows,
	tableInfo._label );

    errstr << "Added table with qualified name '" << table -> qualifiedName() << "'" << Sas::message( Msg::Core, Msg::Verbose );

    RETURN table;
}


FitsReaderWriter::Status
FitsReaderWriter::columnDataTypeStatus( int typeCode ) const
{
    PROC
	Status		status = FitsReaderWriter::Good;

    if( !isValidDalDataType( abs( typeCode ) ))
    {
	status = FitsReaderWriter::Bad;
    } // if

    RETURN status;
}


void
FitsReaderWriter::setColumnDataType( ColumnInfo& columnInfo ) const
{
    PROC
	try {
	    SetExceptionErrorHandler eeh;
	    columnInfo._dataType = fromFITSDataType( abs( columnInfo._typeCode ));
	}
	catch(SasException::Error e){
	    char n[256]; 
	    fits_file_name(columnInfo._fptr, n, &status);
	    errstr << "There was an error while trying to determine the type of column "
		   << columnInfo._name << " in file " << std::string(n)
		   << " The error said: [" << e.code << "]" <<e.message
		   << Sas::error("FITSError");
	}
    ENDPROC
	}


int
FitsReaderWriter::readTable(
    fitsfile * 	fptr,
    DataSet &	dataSet,
    TableInfo&	tableInfo,
    int &		status ) const
{
    PROC

	if( status != 0 )
	{
	    ALTRETURNV( 1 );
	} // if

    //TableInfo	tableInfo;

    // *************************************************
    // Get all the information pertaining to this table.
    // *************************************************
    readTableInfo( fptr, tableInfo, status );

    // ******************************
    // Add the table to the data set.
    // ******************************
    Table*	table = addTable( dataSet, tableInfo );

    // *********************
    // Add table attributes.
    // *********************
    const unsigned int tableAttributes( tableInfo._attributes.size() );

    for( unsigned int i = 0; i < tableAttributes; ++i )
    {
	AttributeInfo	ai( tableInfo._attributes[ i ] );
	Attribute* 	attribute( table -> addAttribute( ai._name, ai._label, ai._units ));

	setAttributeValue( attribute, ai._value );

	errstr << "Added attribute with qualified name '" << attribute -> qualifiedName() << "'" << Sas::message( Msg::Core, Msg::Verbose );
    } // for

    const unsigned int columns( tableInfo._columns.size() );

    for( unsigned int i = 0; i < columns; ++i )
    {
	ColumnInfo	ci( tableInfo._columns[ i ] );
	Column*		column;

	if( columnDataTypeStatus( ci._typeCode ) == FitsReaderWriter::Ignore )
	{
	    std::cout << "Skipped column " << i << std::endl;
	    continue;
	} // if
	//ci._dataType = fromFITSDataType( abs( ci._typeCode ));
	setColumnDataType( ci );
	checkAndSetUnsignedColumn( ci );

	column = table -> addColumn(
	    ci._name,
	    ci._dataType,
	    ci._label,
	    ci._units,
	    ci._size );

	if( ci._nullValueFound )
	{
	    column -> nullValue( ci._nullValue );
	} // if

	errstr << "Added column with qualified name '" << column -> qualifiedName() << "'. type = " << toString( column -> dataType() ) << " rows = " << column -> rows() << " elements = " << column -> elements() << " dimensions = [" << column -> dimensions() << "]" << Sas::message( Msg::Core, Msg::Verbose );

	// *********************
	// Add column attributes.
	// *********************
	const unsigned int columnAttributes( ci._attributes.size() );

	for( unsigned int i = 0; i < columnAttributes; ++i )
	{
	    AttributeInfo	ai( ci._attributes[ i ] );
	    Attribute*	attribute( column -> addAttribute( ai._name, ai._label, ai._units ));

	    setAttributeValue( attribute, ai._value );

	    errstr << "Added attribute with qualified name '" << attribute -> qualifiedName() << "'" << Sas::message( Msg::Core, Msg::Verbose );
	} // for
    } // for

    // ******************************
    // Everything is Ok, so return 0.
    // ******************************
    RETURN 0;

}


bool
FitsReaderWriter::exists( const string& setName )
{
    PROC

	// *******************************************************************
	// This functions checks to see if the named file is a FITS file.
	// There are two valid types for FITS cases to consider:
	// (1) The named file is a FITS file.
	// This case will be tested first as this will be the most common and
	// and also it can be done without using fitsio to opne the fits since
	// we can just look for the FITS 'magic number ' name using the
	// standard c++ io libraries
	// (2) THe named file is a 'zipped' file containing a FITS file.
	// This case is best tested using the cfitsio library to simple open
	// and close the file. This is an expensive operation and is thus
	// tested only if the first test fails.
	// *******************************************************************

	ifstream	ifs( setName.c_str() );


    // *****************************************
    // Check to see if the file actually exists.
    // *****************************************
    if (!ifs)
    {
	ALTRETURNV( false );
    } // if


    // ********
    // CASE (1)
    // ********
    char	s[7];
    string	magic(s);

    ifs.read(s,6);
    s[6] = 0;

    // ****************************************
    // Check whether the magic number is known.
    // ****************************************
    if( magic == FitsReaderWriter::magic() )
    {
	ALTRETURNV( true );

    } // if

    // **************************************************************
    // Clean up before proceeding to second case by closing the file.
    // **************************************************************
    ifs.close();

    // ********
    // CASE (2)
    // ********
    int		errorStatus = 0;
    fitsfile*	fptr = 0;
    bool		found = false;


    // **********************************************
    // Open the FITS Data File.
    // Then check that the file opened without error.
    // **********************************************
    fitsOpenFile(&fptr, setName.c_str(), READONLY, &errorStatus );

    // ************************************************
    // If the file opened successfully the error status
    // will not have been set, and will thus still have
    // it's defualt value (0 in this case).
    // ************************************************
    if( errorStatus == 0 )
    {
	found = true;
    } // if

    // *******************************************************
    // If there was an error in opening the file, then
    // it can't be closed (fptr will not have been
    // set and thus will still have it's default value
    // (which is 0 in this case ), so this needs to be checked
    // before closing.
    // *******************************************************
    if( fptr )
    {
	fitsCloseFile( fptr, &errorStatus );
    } // if
    RETURN found;
}



